@@ -1,7 +1,6 @@
 import asyncio
 
 import engineio
-import six
 
 from . import asyncio_manager
 from . import exceptions
@@ -73,8 +72,7 @@ class AsyncServer(server.Server):
         if client_manager is None:
             client_manager = asyncio_manager.AsyncManager()
         super().__init__(client_manager=client_manager, logger=logger,
-                         binary=False, json=json,
-                         async_handlers=async_handlers, **kwargs)
+                         json=json, async_handlers=async_handlers, **kwargs)
 
     def is_asyncio_based(self):
         return True
@@ -91,8 +89,9 @@ class AsyncServer(server.Server):
                       ``'connect'``, ``'message'`` and ``'disconnect'`` are
                       reserved and should not be used.
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The recipient of the message. This can be set to the
                    session ID of a client to address only that client, or to
                    to any custom room created by the application to address all
@@ -142,8 +141,9 @@ class AsyncServer(server.Server):
         :func:`emit` to issue custom event names.
 
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The recipient of the message. This can be set to the
                    session ID of a client to address only that client, or to
                    to any custom room created by the application to address all
@@ -183,8 +183,9 @@ class AsyncServer(server.Server):
                       ``'connect'``, ``'message'`` and ``'disconnect'`` are
                       reserved and should not be used.
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The session ID of the recipient client.
         :param sid: Alias for the ``to`` parameter.
         :param namespace: The Socket.IO namespace for the event. If this
@@ -209,6 +210,8 @@ class AsyncServer(server.Server):
 
         Note 2: this method is a coroutine.
         """
+        if to is None and sid is None:
+            raise ValueError('Cannot use call() to broadcast.')
         if not self.async_handlers:
             raise RuntimeError(
                 'Cannot use call() when async_handlers is False.')
@@ -224,7 +227,7 @@ class AsyncServer(server.Server):
         try:
             await asyncio.wait_for(callback_event.wait(), timeout)
         except asyncio.TimeoutError:
-            six.raise_from(exceptions.TimeoutError(), None)
+            raise exceptions.TimeoutError() from None
         return callback_args[0] if len(callback_args[0]) > 1 \
             else callback_args[0][0] if len(callback_args[0]) == 1 \
             else None
@@ -326,16 +329,17 @@ class AsyncServer(server.Server):
         Note: this method is a coroutine.
         """
         namespace = namespace or '/'
-        if (ignore_queue and self.manager.is_connected(sid, namespace)) or \
-                await self.manager.can_disconnect(sid, namespace):
+        if ignore_queue:
+            delete_it = self.manager.is_connected(sid, namespace)
+        else:
+            delete_it = await self.manager.can_disconnect(sid, namespace)
+        if delete_it:
             self.logger.info('Disconnecting %s [%s]', sid, namespace)
-            self.manager.pre_disconnect(sid, namespace=namespace)
-            await self._send_packet(sid, packet.Packet(packet.DISCONNECT,
-                                                       namespace=namespace))
+            eio_sid = self.manager.pre_disconnect(sid, namespace=namespace)
+            await self._send_packet(eio_sid, packet.Packet(
+                packet.DISCONNECT, namespace=namespace))
             await self._trigger_event('disconnect', namespace, sid)
             self.manager.disconnect(sid, namespace=namespace)
-            if namespace == '/':
-                await self.eio.disconnect(sid)
 
     async def handle_request(self, *args, **kwargs):
         """Handle an HTTP request from the client.
@@ -387,31 +391,28 @@ class AsyncServer(server.Server):
         else:
             data = []
         await self._send_packet(sid, packet.Packet(
-            packet.EVENT, namespace=namespace, data=[event] + data, id=id,
-            binary=None))
+            packet.EVENT, namespace=namespace, data=[event] + data, id=id))
 
-    async def _send_packet(self, sid, pkt):
+    async def _send_packet(self, eio_sid, pkt):
         """Send a Socket.IO packet to a client."""
         encoded_packet = pkt.encode()
         if isinstance(encoded_packet, list):
-            binary = False
             for ep in encoded_packet:
-                await self.eio.send(sid, ep, binary=binary)
-                binary = True
+                await self.eio.send(eio_sid, ep)
         else:
-            await self.eio.send(sid, encoded_packet, binary=False)
+            await self.eio.send(eio_sid, encoded_packet)
 
-    async def _handle_connect(self, sid, namespace):
+    async def _handle_connect(self, eio_sid, namespace):
         """Handle a client connection request."""
         namespace = namespace or '/'
-        self.manager.connect(sid, namespace)
+        sid = self.manager.connect(eio_sid, namespace)
         if self.always_connect:
-            await self._send_packet(sid, packet.Packet(packet.CONNECT,
-                                                       namespace=namespace))
+            await self._send_packet(eio_sid, packet.Packet(
+                packet.CONNECT, {'sid': sid}, namespace=namespace))
         fail_reason = None
         try:
             success = await self._trigger_event('connect', namespace, sid,
-                                                self.environ[sid])
+                                                self.environ[eio_sid])
         except exceptions.ConnectionRefusedError as exc:
             fail_reason = exc.error_args
             success = False
@@ -419,36 +420,31 @@ class AsyncServer(server.Server):
         if success is False:
             if self.always_connect:
                 self.manager.pre_disconnect(sid, namespace)
-                await self._send_packet(sid, packet.Packet(
+                await self._send_packet(eio_sid, packet.Packet(
                     packet.DISCONNECT, data=fail_reason, namespace=namespace))
+            else:
+                await self._send_packet(eio_sid, packet.Packet(
+                    packet.CONNECT_ERROR, data=fail_reason,
+                    namespace=namespace))
             self.manager.disconnect(sid, namespace)
-            if not self.always_connect:
-                await self._send_packet(sid, packet.Packet(
-                    packet.ERROR, data=fail_reason, namespace=namespace))
-            if sid in self.environ:  # pragma: no cover
-                del self.environ[sid]
         elif not self.always_connect:
-            await self._send_packet(sid, packet.Packet(packet.CONNECT,
-                                                       namespace=namespace))
+            await self._send_packet(eio_sid, packet.Packet(
+                packet.CONNECT, {'sid': sid}, namespace=namespace))
 
-    async def _handle_disconnect(self, sid, namespace):
+    async def _handle_disconnect(self, eio_sid, namespace):
         """Handle a client disconnect."""
         namespace = namespace or '/'
-        if namespace == '/':
-            namespace_list = list(self.manager.get_namespaces())
-        else:
-            namespace_list = [namespace]
-        for n in namespace_list:
-            if n != '/' and self.manager.is_connected(sid, n):
-                await self._trigger_event('disconnect', n, sid)
-                self.manager.disconnect(sid, n)
-        if namespace == '/' and self.manager.is_connected(sid, namespace):
-            await self._trigger_event('disconnect', '/', sid)
-            self.manager.disconnect(sid, '/')
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
+        if not self.manager.is_connected(sid, namespace):  # pragma: no cover
+            return
+        self.manager.pre_disconnect(sid, namespace=namespace)
+        await self._trigger_event('disconnect', namespace, sid)
+        self.manager.disconnect(sid, namespace)
 
-    async def _handle_event(self, sid, namespace, id, data):
+    async def _handle_event(self, eio_sid, namespace, id, data):
         """Handle an incoming client event."""
         namespace = namespace or '/'
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
         self.logger.info('received event "%s" from %s [%s]', data[0], sid,
                          namespace)
         if not self.manager.is_connected(sid, namespace):
@@ -457,11 +453,13 @@ class AsyncServer(server.Server):
             return
         if self.async_handlers:
             self.start_background_task(self._handle_event_internal, self, sid,
-                                       data, namespace, id)
+                                       eio_sid, data, namespace, id)
         else:
-            await self._handle_event_internal(self, sid, data, namespace, id)
+            await self._handle_event_internal(self, sid, eio_sid, data,
+                                              namespace, id)
 
-    async def _handle_event_internal(self, server, sid, data, namespace, id):
+    async def _handle_event_internal(self, server, sid, eio_sid, data,
+                                     namespace, id):
         r = await server._trigger_event(data[0], namespace, sid, *data[1:])
         if id is not None:
             # send ACK packet with the response returned by the handler
@@ -472,16 +470,15 @@ class AsyncServer(server.Server):
                 data = list(r)
             else:
                 data = [r]
-            await server._send_packet(sid, packet.Packet(packet.ACK,
-                                                         namespace=namespace,
-                                                         id=id, data=data,
-                                                         binary=None))
+            await server._send_packet(eio_sid, packet.Packet(
+                packet.ACK, namespace=namespace, id=id, data=data))
 
-    async def _handle_ack(self, sid, namespace, id, data):
+    async def _handle_ack(self, eio_sid, namespace, id, data):
         """Handle ACK packets from the client."""
         namespace = namespace or '/'
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
         self.logger.info('received ack from %s [%s]', sid, namespace)
-        await self.manager.trigger_callback(sid, namespace, id, data)
+        await self.manager.trigger_callback(sid, id, data)
 
     async def _trigger_event(self, event, namespace, *args):
         """Invoke an application event handler."""
@@ -502,49 +499,51 @@ class AsyncServer(server.Server):
             return await self.namespace_handlers[namespace].trigger_event(
                 event, *args)
 
-    async def _handle_eio_connect(self, sid, environ):
+    async def _handle_eio_connect(self, eio_sid, environ):
         """Handle the Engine.IO connection event."""
         if not self.manager_initialized:
             self.manager_initialized = True
             self.manager.initialize()
-        self.environ[sid] = environ
-        return await self._handle_connect(sid, '/')
+        self.environ[eio_sid] = environ
 
-    async def _handle_eio_message(self, sid, data):
+    async def _handle_eio_message(self, eio_sid, data):
         """Dispatch Engine.IO messages."""
-        if sid in self._binary_packet:
-            pkt = self._binary_packet[sid]
+        if eio_sid in self._binary_packet:
+            pkt = self._binary_packet[eio_sid]
             if pkt.add_attachment(data):
-                del self._binary_packet[sid]
+                del self._binary_packet[eio_sid]
                 if pkt.packet_type == packet.BINARY_EVENT:
-                    await self._handle_event(sid, pkt.namespace, pkt.id,
+                    await self._handle_event(eio_sid, pkt.namespace, pkt.id,
                                              pkt.data)
                 else:
-                    await self._handle_ack(sid, pkt.namespace, pkt.id,
+                    await self._handle_ack(eio_sid, pkt.namespace, pkt.id,
                                            pkt.data)
         else:
             pkt = packet.Packet(encoded_packet=data)
             if pkt.packet_type == packet.CONNECT:
-                await self._handle_connect(sid, pkt.namespace)
+                await self._handle_connect(eio_sid, pkt.namespace)
             elif pkt.packet_type == packet.DISCONNECT:
-                await self._handle_disconnect(sid, pkt.namespace)
+                await self._handle_disconnect(eio_sid, pkt.namespace)
             elif pkt.packet_type == packet.EVENT:
-                await self._handle_event(sid, pkt.namespace, pkt.id, pkt.data)
+                await self._handle_event(eio_sid, pkt.namespace, pkt.id,
+                                         pkt.data)
             elif pkt.packet_type == packet.ACK:
-                await self._handle_ack(sid, pkt.namespace, pkt.id, pkt.data)
+                await self._handle_ack(eio_sid, pkt.namespace, pkt.id,
+                                       pkt.data)
             elif pkt.packet_type == packet.BINARY_EVENT or \
                     pkt.packet_type == packet.BINARY_ACK:
-                self._binary_packet[sid] = pkt
-            elif pkt.packet_type == packet.ERROR:
-                raise ValueError('Unexpected ERROR packet.')
+                self._binary_packet[eio_sid] = pkt
+            elif pkt.packet_type == packet.CONNECT_ERROR:
+                raise ValueError('Unexpected CONNECT_ERROR packet.')
             else:
                 raise ValueError('Unknown packet type.')
 
-    async def _handle_eio_disconnect(self, sid):
+    async def _handle_eio_disconnect(self, eio_sid):
         """Handle Engine.IO disconnect event."""
-        await self._handle_disconnect(sid, '/')
-        if sid in self.environ:
-            del self.environ[sid]
+        for n in list(self.manager.get_namespaces()).copy():
+            await self._handle_disconnect(eio_sid, n)
+        if eio_sid in self.environ:
+            del self.environ[eio_sid]
 
     def _engineio_server_class(self):
         return engineio.AsyncServer
